﻿using System;
using System.Drawing;

namespace Collision
{
    class Sphaere
        : StrokeObject
    {
        /// <summary>
        /// Random Geschwindigkeitsvektor für die Kugeln
        /// </summary>
        private static Random Rnd = new Random();

        private static Color[] Colors = { Color.Red, Color.Green, Color.Blue, Color.Yellow, Color.Violet, Color.Indigo };
        private static int ColorIndex = 0;

        /// <summary>
        /// Farbe zum Ausmalen der Kugel
        /// </summary>
        private Brush brush;

        /// <summary>
        /// Masse der Kugel
        /// </summary>
        public float Mass;

        /// <summary>
        /// Radius der Kugel
        /// </summary>
        public float Radius;

        /// <summary>
        /// Beim Test werden immer dieselben Farben generiert
        /// </summary>
        public static void ResetColor()
        {
            Sphaere.ColorIndex = 0;
        }

        /// <summary>
        /// Sphäre zur Mausposition
        /// </summary>
        /// <param name="x"></param>
        /// <param name="y"></param>
        public Sphaere(Point pt)
        {
            this.Location = new Vector(pt.X, pt.Y, 0);
            this.Speed = new Vector();
            this.Mass = 1;
        }

        public Sphaere(float rad, float mass = 1)
        {
            // oben links in der Ecke
            this.Location = new Vector(2 * rad, 2 * rad, 0);
            this.Speed = Sphaere.InitRandomSpeed();

            this.Radius = rad;
            this.Mass = mass;

            this.SetNextBrush();
        }

        /// <summary>
        /// Copy-Konstruktor
        /// </summary>
        /// <param name="other">das zu kopierende Objekt</param>
        public Sphaere(Sphaere other)
        {
            this.Location = new Vector(other.Location);
            this.Speed = new Vector(other.Speed);

            this.Radius = other.Radius;
            this.Mass = other.Mass;

            this.SetNextBrush();
        }

        /// <summary>
        /// Konstruktor für Spezialfälle
        /// </summary>
        /// <param name="loc">Ortsvektor</param>
        /// <param name="sp">Geschwindigkeitsvektor</param>
        public Sphaere(float rad, float mass, Vector loc, Vector sp)
        {
            this.Radius = rad;
            this.Mass = mass;

            this.Location = loc;
            this.Speed = sp;

            this.SetNextBrush();
        }

        /// <summary>
        /// Farbe der aktuellen Kugel bereitstellen
        /// </summary>
        private void SetNextBrush()
        {
            this.brush = new SolidBrush(Sphaere.Colors[Sphaere.ColorIndex]);
            Sphaere.ColorIndex = (Sphaere.ColorIndex + 1) % Sphaere.Colors.Length;
        }

        /// <summary>
        /// Geschwindigkeitsvektor setzen
        /// </summary>
        /// <returns></returns>
        private static Vector InitRandomSpeed()
        {
            Vector speed = new Vector();
            speed.X = (float)Sphaere.Rnd.NextDouble();
            speed.Y = (float)Sphaere.Rnd.NextDouble();
            //Speed.Z = (float)BaseObject.Rnd.NextDouble();
            speed.Z = 0;
            return speed;
        }

        /// <summary>
        /// Testet, ob die gegenwärtige Position
        /// innerhalb des x-y-Spielfeldes liegt.
        /// Ansonsten wird eine Reflexion
        /// an der starren Wand durchgeführt.
        /// </summary>
        /// <param name="border">Spielfeldrand</param>
        /// <returns>True: innerhalb</returns>
        public bool CheckWall(RectangleF border)
        {
            RectangleF pos = new RectangleF(this.Location.X, this.Location.Y, 0, 0);
            pos.Inflate(this.Radius, this.Radius);
            bool crash = !border.Contains(pos);
            if (crash)
            {
                float delta;    // Position, bei der der Crash festgestellt wurde
                float time;     // negative Zeit, bei der der Crash stattgefunden hätte
                if (this.Location.X - this.Radius < 0)
                {
                    // Reflexion links
                    delta = this.Location.X - this.Radius;
                    time = delta / this.Speed.X;

                    this.Step(-time);
                    this.Speed.X = -this.Speed.X;
                    this.Step(time);
                }
                else if (this.Location.X + this.Radius > border.Width)
                {
                    // Reflexion rechts
                    delta = (this.Location.X + this.Radius) - border.Width;
                    time = delta / this.Speed.X;

                    this.Step(-time);
                    this.Speed.X = -this.Speed.X;
                    this.Step(time);
                }

                if (this.Location.Y - this.Radius < 0)
                {
                    // Reflexion oben
                    delta = this.Location.Y - this.Radius;
                    time = delta / this.Speed.Y;

                    this.Step(-time);
                    this.Speed.Y = -this.Speed.Y;
                    this.Step(time);
                }
                else if (this.Location.Y + this.Radius > border.Height)
                {
                    // Reflexion unten
                    delta = (this.Location.Y + this.Radius) - border.Height;
                    time = delta / this.Speed.Y;

                    this.Step(-time);
                    this.Speed.Y = -this.Speed.Y;
                    this.Step(time);
                }
            }

            return crash;
        }

        public static bool CheckCrash(Sphaere b1, Sphaere b2)
        {
            Vector relPos = b2.Location - b1.Location;
            float rSum = b2.Radius + b1.Radius;
            if (rSum < relPos.Length())
            {
                return false;
            }

            Vector relSpeed = b2.Speed - b1.Speed;
            Vector pCollision = Sphaere.VectorCircleEnter(rSum, relSpeed, relPos);
            Vector speed = pCollision - relPos;
            float moveBack = speed.Length();
            float tDelay = moveBack / relSpeed.Length();

            b1.Step(-tDelay);
            b2.Step(-tDelay);

            speed = b1.Speed;
            Vector speed2 = b2.Speed;
            Sphaere.Collision_2D(pCollision, b1.Mass, ref speed, b2.Mass, ref speed2);
            b2.Speed = speed2;
            b1.Speed = speed;

            b1.Step(tDelay);
            b2.Step(tDelay);

            return true;
        }

        /// <summary>
        /// Zweidimensionaler elastischer Stoß,
        /// momentan mit gleicher Masse
        /// </summary>
        /// <param name="grn">Grüne Kugel beim Test</param>
        /// <param name="red">Rote Kugel beim Test</param>
        /// <returns></returns>
        public static bool CheckCrash2(Sphaere grn, Sphaere red)
        {
            // https://de.wikipedia.org/wiki/Sto%C3%9F_(Physik)

            Vector relPos = (red.Location - grn.Location);    // dist
            float distance = relPos.Length();
            float rSum = (grn.Radius + red.Radius);     // EDR
            if (distance >= rSum)
            {
                return false;
            }

            // EDR: Positionskorrektur auf den genauen Zeitpunkt des Stoßes
            Vector relSpeed = red.Speed - grn.Speed;

            Sphaere.Collision_2D(relPos, grn.Mass, ref grn.Speed, red.Mass, ref red.Speed);
            return true;



            float m1 = grn.Mass;
            float m2 = red.Mass;
            float m12 = m1 + m2;

            Vector impulsA1 = m1 * grn.Speed;
            Vector impulsA2 = m2 * red.Speed;

            // Differenzvektor merken
            Vector n = relPos.Clone();
            n.Normalize();

            // parallele Richtungskomponente für Grün
            Vector vnGrn = Sphaere.ParallelComponent(n, grn.Speed);
            // parallele Richtungskomponente für Rot
            Vector vnRed = Sphaere.ParallelComponent(n, red.Speed);

            // senkrechte Richtungskomponenten
            Vector vEGrn = grn.Speed - vnGrn;
            Vector vERed = red.Speed - vnRed;

#if false   // true // ist noch falsch
#if false   //true //
            // dies ist voll falsch
            grn.Speed = m1 * vnGrn + m2 * (2 * vnRed - vnGrn) / m12;
            red.Speed = m2 * vnRed + m1 * (2 * vnGrn - vnRed) / m12;
#else
            // wie beim zentralen Stoß
            vnGrn = 2F * ((m1 * vnGrn) + (m2 * vnRed)) / (m1 + m2) - vnGrn;
            vnRed = 2F * ((m1 * vnGrn) + (m2 * vnRed)) / (m1 + m2) - vnRed;

            //vEGrn = 2F * ((m1 * vEGrn) + (m2 * vERed)) / (m1 + m2) - vEGrn;
            //vERed = 2F * ((m1 * vEGrn) + (m2 * vERed)) / (m1 + m2) - vERed;

            grn.Speed = vnRed + vEGrn;
            red.Speed = vnGrn + vERed;
#endif
#else
            // dies geht
            grn.Speed = vnRed + vEGrn;
            red.Speed = vnGrn + vERed;
#endif

            //Console.WriteLine("Impuls 1: " + impulsA1);
            //Console.WriteLine("Impuls 2: " + impulsA2);
            Vector impulsB1 = m1 * grn.Speed;
            Vector impulsB2 = m2 * red.Speed;
            //Console.WriteLine("Impuls 3: " + impulsB1);
            //Console.WriteLine("Impuls 4: " + impulsB2);

            impulsA1 += impulsA2;
            impulsB1 += impulsB2;
            Vector delta = impulsA1 - impulsB1;
            //Console.WriteLine("Impuls-Betrag: " + delta.Betrag());

            return true;
        }

        /// <summary>
        /// Berechnung der parallelen Richtungskomponenten nach dem Stoß
        /// </summary>
        /// <param name="n">Richtungsvektor</param>
        /// <param name="v">zu zerlegender Komponentenvektor</param>
        /// <returns>die parallele Richtungskomponente</returns>
        private static Vector ParallelComponent(Vector n, Vector v)
        {
            // überprüft
            float factor = (n * v) / n.Length();
            return factor * n;
        }

        /// <summary>
        /// Zeichnen der this Kugel
        /// </summary>
        /// <param name="graphics"></param>
        internal override void Draw(Graphics graphics)
        {
            RectangleF pos = new RectangleF(this.Location.X, this.Location.Y, 0, 0);
            pos.Inflate(this.Radius, this.Radius);
            graphics.FillEllipse(this.brush, pos);
        }

        /// <summary>
        /// Übergang zum nächsten Zeitpunkt
        /// </summary>
        /// <param name="duration"></param>
        internal override void Step(float duration)
        {
            this.Location.ChangePosition(this.Speed, duration);
        }

        /// <summary>
        /// zur Anzeige im Debugger
        /// </summary>
        public override string ToString()
        {
            return string.Format("Speed: {0}, Loc: {1}, Mass: {2}", this.Speed, this.Location, this.Mass);
        }

        /// <summary>
        /// Zweidimensionaler elastischer Stoß zweier Kugeln,
        /// die Kugeln können unterschiedliche Masse haben
        /// </summary>
        /// <param name="positionDelta">Differenzvektor der Masse-Schwerpunkte zum Kollisions-Zeitpunkt</param>
        /// <param name="m1">Masse 1</param>
        /// <param name="speed1">Geschwindigkeit 1</param>
        /// <param name="m2">Masse 2</param>
        /// <param name="speed2">Geschwindigkeit 2</param>
        private static void Collision_2D(Vector positionDelta, float m1, ref Vector speed1, float m2, ref Vector speed2)
        {
            positionDelta.Normalize();

            Vector vColl = Sphaere.ParallelComponent(positionDelta, speed1);
            Vector vColl2 = Sphaere.ParallelComponent(positionDelta, speed2);

            //Vector vColl = v1 * positionDelta * positionDelta;
            //Vector vColl2 = v2 * positionDelta * positionDelta;

            Vector vOrtho = speed1 - vColl;
            Vector vOrtho2 = speed2 - vColl2;
            Sphaere.Collision_1D(m1, ref vColl, m2, ref vColl2);
            speed1 = vOrtho + vColl;
            speed2 = vOrtho2 + vColl2;
        }

        /// <summary>
        /// Eindimensionaler elastischer Stoß zweier Kugeln,
        /// die Kugeln können unterschiedliche Masse haben
        /// </summary>
        /// <param name="m1">Masse 1</param>
        /// <param name="speed1">Geschwindigkeit 1</param>
        /// <param name="m2">Masse 2</param>
        /// <param name="speed2">Geschwindigkeit 2</param>
        private static void Collision_1D(float m1, ref Vector speed1, float m2, ref Vector speed2)
        {
            Vector weightedAverageSpeed = 2F * (m1 * speed1 + m2 * speed2) / (m1 + m2);
            speed1 = weightedAverageSpeed - speed1;
            speed2 = weightedAverageSpeed - speed2;
        }

        /// <summary>
        /// Kreis-Eintrittspunkt des Vectors "direction", dessen Gerade durch ptSupport verläuft - also ein Schnittpunkt der Sekante
        /// </summary>
        /// <param name="radius"></param>
        /// <param name="direction"></param>
        /// <param name="ptSupport"></param>
        /// <returns></returns>
        private static Vector VectorCircleEnter(float radius, Vector direction, Vector ptSupport)
        {
            // https://de.wikipedia.org/wiki/Schnittpunkt#Schnittpunkte_einer_Gerade_mit_einem_Kreis

            Vector directionNormale = new Vector(direction.Y, -direction.X, 0);
            float a = directionNormale.X;
            float b = directionNormale.Y;
            float c = ptSupport * directionNormale;
            float ab2 = a * a + b * b;
            float wurzel = (float)Math.Sqrt(radius * radius * ab2 - c * c);
            float x = (a * c + b * wurzel) / ab2;
            float y = (b * c - a * wurzel) / ab2;
            return new Vector(x, y, 0);
        }

        protected override void Dispose(bool disposing)
        {
            if (disposing)
            {
                if (this.brush != null)
                {
                    this.brush.Dispose();
                }
            }

            base.Dispose(disposing);
        }
    }
}
